import cacahuate.auth.base
import cacahuate.errors
import cacahuate.models

import ldap3
import ldap3.core.exceptions


class LdapAuthProvider(cacahuate.auth.base.BaseAuthProvider):
    def authenticate(self, **credentials):
        if 'username' not in credentials or not credentials['username']:
            raise cacahuate.errors.AuthFieldRequired('username')
        if 'password' not in credentials or not credentials['password']:
            raise cacahuate.errors.AuthFieldRequired('password')

        server_uri = self.config['AUTH_LDAP_SERVER_URI']
        use_ssl = self.config['AUTH_LDAP_USE_SSL']
        domain = self.config['AUTH_LDAP_DOMAIN']
        search_base = self.config['AUTH_LDAP_SEARCH_BASE']
        search_filter = self.config['AUTH_LDAP_SEARCH_FILTER']

        # Use credentials to authenticate
        username = credentials['username'].lower()
        password = credentials['password']
        if 'domain' in credentials:
            domain = credentials['domain']

        # Connect & query ldap
        server = ldap3.Server(
            server_uri,
            get_info=ldap3.ALL,
            use_ssl=use_ssl,
        )

        try:
            conn = ldap3.Connection(
                server,
                user='{}\\{}'.format(domain, username),
                password=password,
                auto_bind=True,
            )
        except ldap3.core.exceptions.LDAPBindError:
            raise cacahuate.errors.AuthFieldInvalid('password')
        except ldap3.core.exceptions.LDAPSocketOpenError:
            raise cacahuate.errors.MisconfiguredProvider(
                f'Can\'t reach {server_uri}',
            )

        conn.search(
            search_base,
            search_filter.format(user=username),
            attributes=list(self.config['AUTH_LDAP_USER_ATTR_MAP'].values()),
        )

        try:
            entry = conn.entries[0]
        except IndexError:
            raise cacahuate.errors.AuthFieldInvalid('username')

        identifier = '{domain}\\{username}'.format(
            domain=domain,
            username=username,
        )

        user = cacahuate.models.get_or_create_user(
            identifier,
            {
                k: getattr(entry, v)
                for k, v in self.config['AUTH_LDAP_USER_ATTR_MAP'].items()
            },
        )

        # update if required
        for k, v in self.config['AUTH_LDAP_USER_ATTR_MAP'].items():
            setattr(user, k, getattr(entry, v))
        user.save()

        return (identifier, {
            'identifier': user.identifier,
            'email': user.email,
            'fullname': user.fullname,
        })



#!/usr/bin/env python3
from collections import ChainMap
from xml.dom import pulldom
from xml.sax._exceptions import SAXParseException
import argparse
import logging
import logging.config
import os
import re
import sys
import time
from importlib import import_module

from coralillo import Engine
from itacate import Config
from lark.exceptions import GrammarError, ParseError, LexError

from cacahuate.errors import MalformedProcess
from cacahuate.grammar import Condition
from cacahuate.indexes import create_indexes
from cacahuate.models import bind_models
from cacahuate.xml import NODES, get_text
from cacahuate.tasks import app as celery


def main():
    # Load the config
    config = Config(os.path.dirname(os.path.realpath(__file__)))
    config.from_object('cacahuate.settings')

    if os.getenv('CACAHUATE_SETTINGS'):
        config.from_envvar('CACAHUATE_SETTINGS', silent=False)

    # Set the timezone
    os.environ['TZ'] = config['TIMEZONE']
    time.tzset()

    # Setup logging
    logging.config.dictConfig(config['LOGGING'])

    # Load the models
    eng = Engine(
        host=config['REDIS_HOST'],
        port=config['REDIS_PORT'],
        db=config['REDIS_DB'],
        id_function=getattr(
            import_module(config['DB_ID_FUNCTION'].rsplit('.', 1)[0]),
            config['DB_ID_FUNCTION'].rsplit('.', 1)[1],
        ),
    )
    bind_models(eng)

    # Create mongo indexes
    create_indexes(config)

    # Update celery config
    celery.conf.update(
        broker='amqp://{user}@{host}//'.format(
            user=config['RABBIT_USER'],
            host=config['RABBIT_HOST'],
        ),
        worker_concurrency=1,
        worker_hijack_root_logger=False,
        task_default_queue=config['RABBIT_QUEUE'],
    )
    worker = celery.Worker()
    worker.start()


def rng_path():
    print(os.path.abspath(os.path.join(
        os.path.dirname(__file__),
        'xml/process-spec.rng'
    )))


def _validate_file(filename, verbose=False):
    ids = []
    data_form = ChainMap()
    passed_nodes = []
    variable_re = re.compile(r'^[a-zA-Z_][a-zA-Z0-9_]*$')

    class StreamWrapper:

        def __init__(self, filename):
            self.f = open(filename)
            self.lineno = 0

        def read(self, b):
            self.lineno += 1
            return self.f.readline()

    sw = StreamWrapper(filename)

    def check_id(node):
        id_element = node.getAttribute('id')

        if not id_element:
            raise MalformedProcess(
                '{}:{} All nodes must have an id'.format(filename, sw.lineno)
            )

        if not variable_re.match(id_element):
            raise MalformedProcess(
                '{}:{} Id must be a valid variable name'.format(
                    filename, sw.lineno,
                )
            )

        if id_element not in ids:
            ids.append(id_element)
        else:
            raise MalformedProcess("{}:{} Duplicated id: '{}'".format(
                filename, sw.lineno,
                id_element,
            ))

    doc = pulldom.parse(sw)

    for event, node in doc:
        if event == pulldom.START_ELEMENT and node.tagName == 'block':
            data_form.maps.append(dict())

        if event == pulldom.END_ELEMENT and node.tagName == 'block':
            data_form.maps.pop()

        if event != pulldom.START_ELEMENT:
            continue

        if node.tagName == 'condition':
            doc.expandNode(node)

            try:
                tree = Condition().parse(get_text(node))
            except GrammarError:
                raise MalformedProcess(
                    '{}:{} Grammar error in condition'.format(
                        filename, sw.lineno,
                    )
                )
            except ParseError:
                raise MalformedProcess(
                    '{}:{} Parse error in condition'.format(
                        filename, sw.lineno,
                    )
                )
            except LexError:
                raise MalformedProcess(
                    '{}:{} Lex error in condition'.format(filename, sw.lineno)
                )
            except KeyError as e:
                raise MalformedProcess(
                    '{}:{} variable does not exist: {}'.format(
                        filename, sw.lineno,
                        str(e),
                    )
                )

            # validate variables used in condition
            for tree in tree.find_data('ref'):
                reference_form, field_form = list(map(
                    lambda x: x.children[0][:],
                    tree.children
                ))

                try:
                    data_form[reference_form][field_form]
                except KeyError:
                    raise MalformedProcess(
                        "{}:{} variable used in condition "
                        "is not defined '{}'".format(
                            filename, sw.lineno,
                            reference_form + '.' + field_form,
                        )
                    )

            continue

        if node.tagName not in NODES:
            continue

        # Duplicate ids
        check_id(node)

        if node.tagName == 'if':
            continue

        # Expand and check this node. <if> nodes are not expanded
        doc.expandNode(node)

        # Check auth-filter params
        params = node.getElementsByTagName("param")

        for param in params:
            if param.getAttribute('type') != 'ref':
                continue

            ref_type, ref = get_text(param).split('#')

            if ref_type == 'form':
                reference_form, field_form = ref.split('.')

                try:
                    data_form[reference_form][field_form]
                except KeyError:
                    raise MalformedProcess(
                        "{}:{} Referenced param does not exist '{}'".format(
                            filename, sw.lineno,
                            reference_form + '.' + field_form,
                        )
                    )
            elif ref_type == 'user':
                if ref not in passed_nodes:
                    raise MalformedProcess(
                        '{}:{} Referenced user is never created: {}'.format(
                            filename, sw.lineno,
                            ref,
                        )
                    )

        # Check dependencies
        deps = node.getElementsByTagName("dep")

        for dep in deps:
            form, field = get_text(dep).split('.')

            try:
                data_form[form][field]
            except KeyError:
                raise MalformedProcess(
                    "{}:{} Referenced dependency does not exist '{}'".format(
                        filename, sw.lineno,
                        form + '.' + field,
                    )
                )

        # fill forms for later usage
        forms = node.getElementsByTagName('form')

        for form in forms:
            form_id = form.getAttribute('id')

            if form_id and not variable_re.match(form_id):
                raise MalformedProcess(
                    '{}:{} Form ids must be valid variable names'.format(
                        filename, sw.lineno,
                    )
                )

            form_ref = form.getAttribute('ref')

            if form_ref and not variable_re.match(form_ref):
                raise MalformedProcess(
                    '{}:{} Form refs must be valid variable names'.format(
                        filename, sw.lineno,
                    )
                )

            inputs = form.getElementsByTagName("input")
            array_input = {}

            for inpt in inputs:
                inpt_name = inpt.getAttribute('name')

                if not variable_re.match(inpt_name):
                    raise MalformedProcess(
                        '{}:{} Field names must match [a-zA-Z0-9_]+'.format(
                            filename, sw.lineno,
                        )
                    )

                array_input[inpt_name] = \
                    inpt.getAttribute('default')

            data_form.maps[-1][form_id] = array_input

        # In case of a request node, add its captured values to `data_form`
        # so they can be recognized as valid values
        captures = node.getElementsByTagName('capture')

        for capture in captures:
            capture_id = capture.getAttribute('id')

            if capture_id and not variable_re.match(capture_id):
                raise MalformedProcess(
                    '{}:{} Capture ids must be valid variable names'.format(
                        filename, sw.lineno,
                    )
                )

            capture_inputs = capture.getElementsByTagName("value")
            array_input = {}

            for inpt in capture_inputs:
                inpt_name = inpt.getAttribute('name')

                if not variable_re.match(inpt_name):
                    raise MalformedProcess(
                        '{}:{} names must be valid variable names'.format(
                            filename, sw.lineno,
                        )
                    )

                array_input[inpt_name] = '1'

            data_form.maps[-1][capture_id] = array_input

        # add this node to the list of revised nodes
        has_auth_filter = len(node.getElementsByTagName('auth-filter')) > 0

        if has_auth_filter:
            passed_nodes.append(node.getAttribute('id'))

    if verbose:
        print('{} seems correct'.format(filename))


def xml_validate(filenames=None, verbose=False):
    if not filenames:
        parser = argparse.ArgumentParser(description='Validate xmls')

        parser.add_argument('files', metavar='FILE', nargs='+',
                            help='the files to be validated')
        parser.add_argument('--verbose', '-v', action='store_true')

        args = parser.parse_args()

        filenames = args.files
        verbose = args.verbose

    found_errors = False

    for filename in filenames:
        try:
            try:
                _validate_file(filename, verbose)
            except SAXParseException:
                raise MalformedProcess('{}:{} Is not valid xml'.format(
                    filename, 0
                ))
            except FileNotFoundError:
                raise MalformedProcess('{} not found'.format(filename))
        except MalformedProcess as e:
            found_errors = True
            print(e)

    if found_errors:
        sys.exit('** Validation errors found **')


if __name__ == '__main__':
    main()



from os import path

from coralillo.errors import ModelNotFoundError

import flask

from jinja2 import Environment, FileSystemLoader, select_autoescape

from cacahuate.http.mongo import mongo
from cacahuate.mongo import make_context


bp = flask.Blueprint('template', __name__)


def datetimeformat(value, format='%Y-%m-%d %H:%M:%S%z'):
    return value.strftime(format)


@bp.route('/execution/<id>/summary', methods=['GET'])
def execution_template(id):
    # load values
    collection = mongo.db[flask.current_app.config['EXECUTION_COLLECTION']]

    try:
        execution = next(collection.find({'id': id}))
    except StopIteration:
        raise ModelNotFoundError(
            'Specified execution never existed, and never will'
        )

    if 'process_name' not in execution:
        return 'Not supported for old processes', 409

    context = make_context(execution, flask.current_app.config)

    # load template
    process_name = execution['process_name']
    name, version, _ = process_name.split('.')

    # Loaders will be inserted in inverse order and then reversed. The fallback
    # is the default template at ``templates/summary.html``
    paths = [
        path.join(path.dirname(path.realpath(__file__)), '../../templates'),
    ]

    app_template_path = flask.current_app.config['TEMPLATE_PATH']
    if app_template_path is not None and path.isdir(app_template_path):
        paths.append(app_template_path)

        process_dir = path.join(app_template_path, name)
        if path.isdir(process_dir):
            paths.append(process_dir)

        process_version_dir = path.join(
            process_dir, version
        )
        if path.isdir(process_version_dir):
            paths.append(process_version_dir)

    env = Environment(
        loader=FileSystemLoader(reversed(paths)),
        autoescape=select_autoescape(['html', 'xml'])
    )

    env.filters['datetimeformat'] = datetimeformat

    for name, function in flask.current_app.config['JINJA_FILTERS'].items():
        env.filters[name] = function

    return flask.make_response(
        env.get_template('summary.html').render(**context),
        200,
    )



import pymongo


def create_indexes(config):
    # Create index
    uri = config['MONGO_URI']

    with pymongo.MongoClient(uri) as client:
        database_name = pymongo.uri_parser.parse_uri(uri)['database']
        db = client[database_name]

        db.execution.create_index("id", unique=True)
        db.execution.create_index("status")
        db.execution.create_index("started_at")
        db.execution.create_index("finished_at")

        db.pointer.create_index("status")
        db.pointer.create_index("execution.id")
        db.pointer.create_index("started_at")
        db.pointer.create_index("finished_at")



from case_conversion import pascalcase
from datetime import datetime
from functools import reduce
from operator import and_
import ast
import re
import numbers

from jsonpath_rw import parse as jsonpathparse

from cacahuate.errors import InvalidInputError, RequiredListError
from cacahuate.errors import MisconfiguredProvider, RequiredIntError
from cacahuate.errors import RequiredFloatError, RequiredStrError
from cacahuate.errors import RequiredInputError, InvalidDateError
from cacahuate.xml import get_text

INPUTS = [
    'text',
    'password',
    'checkbox',
    'radio',
    'file',
    'datetime',
    'date',
    'select',
    'int',
    'float',
    'link',
    'currency',
]


class Option:
    def __init__(self, label, value):
        self.label = label
        self.value = value

    def to_json(self):
        return {
            'label': self.label,
            'value': self.value,
        }


class Input:
    """docstring for Input"""

    def __init__(self, element, *args, **kwargs):
        self.type = element.getAttribute('type')
        self.required = element.getAttribute('required') == 'required'
        self.name = element.getAttribute('name')
        self.default = element.getAttribute('default') or None
        self.label = element.getAttribute('label') or self.name
        self.hidden = element.getAttribute('hidden') == 'true'

        self.dependencies = []

        deps_el = element.getElementsByTagName('dependencies')

        if deps_el:
            for dep_el in deps_el[0].getElementsByTagName('dep'):
                self.dependencies.append(get_text(dep_el))

    def validate(self, value, form_index):
        value = value or self.get_default()

        if self.required and (value == '' or value is None):
            raise RequiredInputError(
                self.name,
                'request.body.form_array.{}.{}'.format(form_index, self.name)
            )

        return value

    def get_default(self):
        return self.default

    def make_caption(self, value):
        return value

    def to_json(self):
        return {
            'type': self.type,
            'name': self.name,
            'label': self.label,
            'value': None,
            'value_caption': None,  # This is computed in node.py
            'hidden': self.hidden,
        }


class TextInput(Input):
    pass


class PasswordInput(TextInput):

    def make_caption(self, value):
        return '******'


class IntInput(Input):

    def get_default(self):
        try:
            return int(self.default)
        except ValueError:
            return 0
        except TypeError:
            return 0

    def validate(self, value, form_index):
        if self.required and type(value) != int and not value:
            raise RequiredInputError(
                self.name,
                'request.body.form_array.{}.{}'.format(form_index, self.name)
            )

        if type(value) != int and not value:
            value = self.get_default()

        try:
            value = int(value)
        except ValueError:
            raise RequiredIntError(
                self.name,
                'request.body.form_array.{}.{}'.format(form_index, self.name)
            )

        return value

    def make_caption(self, value):
        return str(value)


class FloatInput(Input):

    def get_default(self):
        if not self.default:
            return 0

        try:
            return float(self.default)
        except ValueError:
            return 0

    def validate(self, value, form_index):
        # allowed types are None, String, and Numbers
        curated = None

        if value is None or value == '':
            curated = None
        elif isinstance(value, numbers.Number) or type(value) == str:
            try:
                curated = float(value)
            except ValueError:
                raise RequiredFloatError(
                    self.name,
                    'request.body.form_array.{}.{}'.format(
                        form_index,
                        self.name
                    )
                )
        else:
            raise RequiredFloatError(
                self.name,
                'request.body.form_array.{}.{}'.format(form_index, self.name)
            )

        if curated is None and self.required:
            raise RequiredInputError(
                self.name,
                'request.body.form_array.{}.{}'.format(form_index, self.name)
            )

        if curated is None:
            curated = self.get_default()

        return curated

    def make_caption(self, value):
        return str(value)


class CurrencyInput(FloatInput):
    pass


class FiniteOptionInput(Input):
    def __init__(self, element, context=None, *args, **kwargs):
        super().__init__(element, *args, **kwargs)
        if not context:
            context = {}

        self.options = []

        opttag = element.getElementsByTagName('options')

        if len(opttag) > 0:
            for opt in opttag[0].getElementsByTagName('option'):
                ref_attr = opt.getAttribute('ref')
                if ref_attr:
                    ref_type, ref_path = ref_attr.split('#')
                    if ref_type == 'form':
                        if not len(jsonpathparse(ref_path).find(context)):
                            continue

                        label_path = opt.getAttribute('label')
                        value_path = opt.getAttribute('value')

                        match = jsonpathparse(ref_path).find(context)[0].value.all()
                        for localdata in match:
                            self.options.append(Option(
                                jsonpathparse(label_path).find(localdata)[0].value,
                                jsonpathparse(value_path).find(localdata)[0].value,
                            ))

                else:
                    self.options.append(Option(
                        opt.getAttribute('label') or get_text(opt),
                        opt.getAttribute('value'),
                    ))

    def __contains__(self, item):
        for opt in self.options:
            if opt.value == item:
                return True

        return False

    def make_caption(self, value):
        for opt in self.options:
            if opt.value == value:
                return opt.label


class CheckboxInput(FiniteOptionInput):

    def validate(self, value, form_index):
        super().validate(value, form_index)

        if value is None:
            value = []

        if type(value) == str:
            try:
                value = ast.literal_eval(value)
            except SyntaxError:
                value = []

        if type(value) is not list:
            raise RequiredListError(
                self.name,
                'request.body.form_array.{}.{}'.format(form_index, value)
            )

        for val in value:
            if val not in self:
                raise InvalidInputError(
                    self.name,
                    'request.body.form_array.{}.{}'.format(
                        form_index,
                        self.name
                    )
                )

        return value

    def make_caption(self, value):
        def find_cap(val):
            for opt in self.options:
                if opt.value == val:
                    return opt.label

        return ', '.join(find_cap(val) for val in value)


class RadioInput(FiniteOptionInput):

    def validate(self, value, form_index):
        super().validate(value, form_index)

        curated = None

        if type(value) is str:
            curated = value

        if curated not in self:
            curated = None

        if self.required and not curated:
            raise RequiredStrError(
                self.name,
                'request.body.form_array.{}.{}'.format(form_index, self.name)
            )

        return curated


class SelectInput(RadioInput):
    pass


class FileInput(Input):

    def __init__(self, element, *args, **kwargs):
        super().__init__(element, *args, **kwargs)

        self.provider = element.getAttribute('provider')

    def validate(self, value, form_index):
        super().validate(value, form_index)

        if not value:
            value = None

        if value is not None and type(value) is not dict:
            raise InvalidInputError(
                self.name,
                'request.body.form_array.{}.{}'.format(form_index, self.name)
            )

        provider = self.provider

        if not value and self.required:
            raise RequiredInputError(
                self.name,
                'request.body.form_array.{}.{}'.format(form_index, self.name)
            )

        if provider == 'doqer':
            if value is not None:
                valid = reduce(
                    and_,
                    map(
                        lambda attr: attr in value and value[attr] is not None,
                        ['id', 'mime', 'name', 'type']
                    )
                )

                if not valid:
                    raise InvalidInputError(
                        self.name,
                        'request.body.form_array.{}.{}'.format(
                            form_index,
                            self.name
                        )
                    )
        else:
            raise MisconfiguredProvider(
                'File provider `{}` not implemented'.format(provider)
            )

        return value


class DatetimeInput(Input):

    def validate(self, value, form_index):
        super().validate(value, form_index)

        value = value or self.get_default()

        if not value:
            if self.required:
                raise RequiredInputError(
                    self.name,
                    'request.body.form_array.{}.{}'.format(
                        form_index,
                        self.name
                    )
                )

            return None

        try:
            datetime.strptime(value, "%Y-%m-%dT%H:%M:%S.%fZ")
        except TypeError:
            raise InvalidDateError(
                self.name,
                'request.body.form_array.{}.{}'.format(form_index, self.name)
            )
        except ValueError:
            raise InvalidDateError(
                self.name,
                'request.body.form_array.{}.{}'.format(form_index, self.name)
            )

        return value

    def get_default(self):
        if self.default == 'now':
            return datetime.now().isoformat() + 'Z'

        return None

    def make_caption(self, value):
        if not value:
            return

        return datetime.strptime(value, '%Y-%m-%dT%H:%M:%S.%fZ').strftime('%c')


class DateInput(DatetimeInput):
    pass


class LinkInput(Input):

    def validate(self, value, form_index):
        super().validate(value, form_index)

        curated = None

        if not value:
            curated = None
        elif isinstance(value, dict):
            if 'label' not in value or 'href' not in value:
                raise InvalidInputError(
                    self.name,
                    'request.body.form_array.{}.{}'.format(
                        form_index,
                        self.name
                    )
                )

            label = value['label']
            href = value['href']

            link_valid = re.match('^(https?://)[a-z0-9.]+/?$', href)

            if not link_valid:
                raise InvalidInputError(
                    self.name,
                    'request.body.form_array.{}.{}'.format(
                        form_index,
                        self.name
                    )
                )

            curated = {
                'label': label,
                'href': href,
            }
        else:
            raise InvalidInputError(
                self.name,
                'request.body.form_array.{}.{}'.format(
                    form_index,
                    self.name
                )
            )

        if not curated and self.required:
            raise RequiredInputError(
                self.name,
                'request.body.form_array.{}.{}'.format(form_index, self.name)
            )

        return curated


def make_input(element, context=None):
    ''' returns a build Input object given an Element object '''
    classattr = element.getAttribute('type')

    if not context:
        context = {}

    if classattr not in INPUTS:
        raise ValueError(
            'Class definition not found for input: {}'.format(classattr)
        )

    class_name = pascalcase(classattr) + 'Input'
    available_classes = __import__(__name__).inputs

    return getattr(available_classes, class_name)(element, context)



import os
import logging
import traceback
from itacate import Config
from celery import Celery

from .handler import Handler

LOGGER = logging.getLogger(__name__)

config = Config(os.path.dirname(os.path.realpath(__file__)))
config.from_object('cacahuate.settings')

if os.getenv('CACAHUATE_SETTINGS'):
    config.from_envvar('CACAHUATE_SETTINGS', silent=False)

handler = Handler(config)
app = Celery()


@app.task(ignore_result=True)
def handle(body):
    try:
        handler(body)
    except Exception:
        LOGGER.error(traceback.format_exc())
